@@ -3,7 +3,7 @@ import logging
 import os
 from decimal import Decimal
 from enum import Enum
-from typing import Dict, List, Tuple, Optional, Any
+from typing import Dict, List, Tuple, Optional, Any, Union
 
 from basemodels import Manifest
 from eth_keys import keys
@@ -30,27 +30,13 @@ GAS_LIMIT = int(os.getenv("GAS_LIMIT", 4712388))
 # Explicit env variable that will use s3 for storing results.
 
 LOG = logging.getLogger("hmt_escrow.job")
+
 Status = Enum("Status", "Launched Pending Partial Paid Complete Cancelled")
 
 
 def status(escrow_contract: Contract, gas_payer: str, gas: int = GAS_LIMIT) -> Enum:
     """Returns the status of the Job.
 
-    >>> from test.hmt_escrow.utils import manifest
-    >>> credentials = {
-    ... 	"gas_payer": "0x1413862C2B7054CDbfdc181B83962CB0FC11fD92",
-    ... 	"gas_payer_priv": "28e516f1e2f99e96a48a23cea1f94ee5f073403a1c68e818263f0eb898f1c8e5"
-    ... }
-    >>> rep_oracle_pub_key = b"2dbc2c2c86052702e7c219339514b2e8bd4687ba1236c478ad41b43330b08488c12c8c1797aa181f3a4596a1bd8a0c18344ea44d6655f61fa73e56e743f79e0d"
-    >>> job = Job(credentials, manifest)
-
-    After deployment status is "Launched".
-
-    >>> job.launch(rep_oracle_pub_key)
-    True
-    >>> status(job.job_contract, job.gas_payer)
-    <Status.Launched: 1>
-
     Args:
         escrow_contract (Contract): the escrow contract of the Job.
         gas_payer (str): an ethereum address paying for the gas costs.
@@ -71,20 +57,6 @@ def manifest_url(
 ) -> str:
     """Retrieves the deployed manifest url uploaded on Job initialization.
 
-    >>> from test.hmt_escrow.utils import manifest
-    >>> credentials = {
-    ... 	"gas_payer": "0x1413862C2B7054CDbfdc181B83962CB0FC11fD92",
-    ... 	"gas_payer_priv": "28e516f1e2f99e96a48a23cea1f94ee5f073403a1c68e818263f0eb898f1c8e5"
-    ... }
-    >>> rep_oracle_pub_key = b"2dbc2c2c86052702e7c219339514b2e8bd4687ba1236c478ad41b43330b08488c12c8c1797aa181f3a4596a1bd8a0c18344ea44d6655f61fa73e56e743f79e0d"
-    >>> job = Job(credentials, manifest)
-    >>> job.launch(rep_oracle_pub_key)
-    True
-    >>> job.setup()
-    True
-    >>> manifest_hash(job.job_contract, job.gas_payer) == job.manifest_hash
-    True
-
     Args:
         escrow_contract (Contract): the escrow contract of the Job.
         gas_payer (str): an ethereum address paying for the gas costs.
@@ -104,20 +76,6 @@ def manifest_hash(
 ) -> str:
     """Retrieves the deployed manifest hash uploaded on Job initialization.
 
-    >>> credentials = {
-    ... 	"gas_payer": "0x1413862C2B7054CDbfdc181B83962CB0FC11fD92",
-    ... 	"gas_payer_priv": "28e516f1e2f99e96a48a23cea1f94ee5f073403a1c68e818263f0eb898f1c8e5"
-    ... }
-    >>> rep_oracle_pub_key = b"2dbc2c2c86052702e7c219339514b2e8bd4687ba1236c478ad41b43330b08488c12c8c1797aa181f3a4596a1bd8a0c18344ea44d6655f61fa73e56e743f79e0d"
-    >>> from test.hmt_escrow.utils import manifest
-    >>> job = Job(credentials, manifest)
-    >>> job.launch(rep_oracle_pub_key)
-    True
-    >>> job.setup()
-    True
-    >>> manifest_hash(job.job_contract, job.gas_payer) == job.manifest_hash
-    True
-
     Args:
         escrow_contract (Contract): the escrow contract of the Job.
         gas_payer (str): an ethereum address paying for the gas costs.
@@ -188,6 +146,9 @@ class Job:
         escrow_addr: str = None,
         multi_credentials: List[Tuple] = [],
         retry: Retry = None,
+        hmt_server_addr: str = None,
+        hmtoken_addr: str = None,
+        gas_limit: int = GAS_LIMIT,
     ):
         """Initializes a Job instance with values from a Manifest class and
         checks that the provided credentials are valid. An optional factory
@@ -280,6 +241,9 @@ class Job:
         self.gas_payer = Web3.toChecksumAddress(credentials["gas_payer"])
         self.gas_payer_priv = credentials["gas_payer_priv"]
         self.multi_credentials = self._validate_multi_credentials(multi_credentials)
+        self.hmt_server_addr = hmt_server_addr
+        self.hmtoken_addr = HMTOKEN_ADDR if hmtoken_addr is None else hmtoken_addr
+        self.gas = gas_limit
 
         # Initialize a new Job.
         if not escrow_addr and escrow_manifest:
@@ -290,7 +254,7 @@ class Job:
         elif escrow_addr and factory_addr and not escrow_manifest:
             if not self._factory_contains_escrow(escrow_addr, factory_addr):
                 raise ValueError(
-                    "Given factory address doesn't contain the given escrow address."
+                    "Given factory address doesn't contain the given escrow" " address."
                 )
             self._access_job(factory_addr, escrow_addr, **credentials)
 
@@ -357,14 +321,14 @@ class Job:
 
         job_addr = self._last_escrow_addr()
         LOG.info("Job's escrow contract deployed to:{}".format(job_addr))
-        self.job_contract = get_escrow(job_addr)
+        self.job_contract = get_escrow(job_addr, self.hmt_server_addr)
 
         (hash_, manifest_url) = upload(self.serialized_manifest, pub_key)
         self.manifest_url = manifest_url
         self.manifest_hash = hash_
         return self.status() == Status.Launched and self.balance() == 0
 
-    def setup(self, gas: int = GAS_LIMIT, sender: str = None) -> bool:
+    def setup(self, sender: str = None) -> bool:
         """Sets the escrow contract to be ready to receive answers from the Recording Oracle.
         The contract needs to be deployed and funded first.
 
@@ -411,7 +375,7 @@ class Job:
         reputation_oracle = str(self.serialized_manifest["reputation_oracle_addr"])
         recording_oracle = str(self.serialized_manifest["recording_oracle_addr"])
         hmt_amount = int(self.amount * 10 ** 18)
-        hmtoken_contract = get_hmtoken()
+        hmtoken_contract = get_hmtoken(self.hmtoken_addr, self.hmt_server_addr)
 
         hmt_transferred = False
         contract_is_setup = False
@@ -420,7 +384,8 @@ class Job:
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
         if sender:
             txn_func = hmtoken_contract.functions.transferFrom
@@ -429,7 +394,9 @@ class Job:
             txn_func = hmtoken_contract.functions.transfer
             func_args = [self.job_contract.address, hmt_amount]
 
-        balance = utils.get_hmt_balance(self.gas_payer, HMTOKEN_ADDR, get_w3())
+        balance = utils.get_hmt_balance(
+            self.gas_payer, self.hmtoken_addr, get_w3(self.hmt_server_addr)
+        )
 
         # make sure there is enough HMT to fund the escrow
         if balance > hmt_amount:
@@ -486,8 +453,9 @@ class Job:
             str(self.status()) == str(Status.Pending) and self.balance() == hmt_amount
         )
 
-    def add_trusted_handlers(self, handlers: List[str], gas: int = GAS_LIMIT) -> bool:
-        """Add trusted handlers that can freely transact with the contract and perform aborts and cancels for example.
+    def add_trusted_handlers(self, handlers: List[str]) -> bool:
+        """Add trusted handlers that can freely transact with the contract and
+         perform aborts and cancels for example.
 
         >>> from test.hmt_escrow.utils import manifest
         >>> credentials = {
@@ -524,7 +492,8 @@ class Job:
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
         func_args = [handlers]
 
@@ -549,7 +518,7 @@ class Job:
         payouts: List[Tuple[str, Decimal]],
         results: Dict,
         pub_key: bytes,
-        gas: int = GAS_LIMIT,
+        encrypt_final_results: bool = True,
     ) -> bool:
         """Performs a payout to multiple ethereum addresses. When the payout happens,
         final results are uploaded to IPFS and contract's state is updated to Partial or Paid
@@ -607,6 +576,8 @@ class Job:
             payouts (List[Tuple[str, int]]): a list of tuples with ethereum addresses and amounts.
             results (Dict): the final answer results stored by the Reputation Oracle.
             pub_key (bytes): the public key of the Reputation Oracle.
+            gas (int): maximum amount of gas the caller is ready to pay.
+            encrypt_final_results (bool): Whether final results must be encrypted.
 
         Returns:
             bool: returns True if paying to ethereum addresses and oracles succeeds.
@@ -617,19 +588,28 @@ class Job:
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
 
-        (hash_, url) = upload(results, pub_key)
-        eth_addrs = [eth_addr for eth_addr, amount in payouts]
-        hmt_amounts = [int(amount * 10 ** 18) for eth_addr, amount in payouts]
+        (hash_, url) = upload(
+            msg=results, public_key=pub_key, encrypt_data=encrypt_final_results
+        )
+
+        eth_addrs = list()
+        hmt_amounts = list()
+
+        for eth_addr, amount in payouts:
+            eth_addrs.append(eth_addr)
+            hmt_amounts.append(int(amount * 10 ** 18))
 
         func_args = [eth_addrs, hmt_amounts, url, hash_, 1]
         try:
             handle_transaction_with_retry(txn_func, self.retry, *func_args, **txn_info)
-            return self._bulk_paid() == True
+            return self._bulk_paid() is True
+
         except Exception as e:
-            LOG.debug(
+            LOG.warn(
                 f"{txn_event} failed with main credentials: {self.gas_payer}, {self.gas_payer_priv} due to {e}. Using secondary ones..."
             )
 
@@ -640,15 +620,12 @@ class Job:
         if not bulk_paid:
             LOG.warning(f"{txn_event} failed with all credentials.")
 
-        return bulk_paid == True
+        return bulk_paid is True
 
-    def abort(self, gas: int = GAS_LIMIT) -> bool:
+    def abort(self) -> bool:
         """Kills the contract and returns the HMT back to the gas payer.
         The contract cannot be aborted if the contract is in Partial, Paid or Complete state.
 
-        Returns:
-            bool: returns True if contract has been destroyed successfully.
-
         >>> credentials = {
         ... 	"gas_payer": "0x1413862C2B7054CDbfdc181B83962CB0FC11fD92",
         ... 	"gas_payer_priv": "28e516f1e2f99e96a48a23cea1f94ee5f073403a1c68e818263f0eb898f1c8e5"
@@ -717,13 +694,14 @@ class Job:
             bool: returns True if contract has been destroyed successfully.
 
         """
-        w3 = get_w3()
+        w3 = get_w3(self.hmt_server_addr)
         txn_event = "Job abortion"
         txn_func = self.job_contract.functions.abort
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
 
         try:
@@ -742,7 +720,7 @@ class Job:
 
         return w3.eth.getCode(self.job_contract.address) == b""
 
-    def cancel(self, gas: int = GAS_LIMIT) -> bool:
+    def cancel(self) -> bool:
         """Returns the HMT back to the gas payer. It's the softer version of abort as the contract is not destroyed.
 
         >>> credentials = {
@@ -801,7 +779,8 @@ class Job:
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
 
         try:
@@ -821,9 +800,7 @@ class Job:
 
         return self.status() == Status.Cancelled
 
-    def store_intermediate_results(
-        self, results: Dict, pub_key: bytes, gas: int = GAS_LIMIT
-    ) -> bool:
+    def store_intermediate_results(self, results: Dict, pub_key: bytes) -> bool:
         """Recording Oracle stores intermediate results with Reputation Oracle's public key to S3
         and updates the contract's state.
 
@@ -885,7 +862,8 @@ class Job:
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
         (hash_, url) = upload(results, pub_key)
         func_args = [url, hash_]
@@ -911,12 +889,7 @@ class Job:
         return results_stored
 
     def complete(
-        self,
-        gas: int = GAS_LIMIT,
-        blocking: bool = False,
-        retries: int = 3,
-        delay: int = 5,
-        backoff: int = 2,
+        self, blocking: bool = False, retries: int = 3, delay: int = 5, backoff: int = 2
     ) -> bool:
         """Completes the Job if it has been paid.
 
@@ -961,7 +934,8 @@ class Job:
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
 
         try:
@@ -981,7 +955,7 @@ class Job:
 
         return self.status() == Status.Complete
 
-    def status(self, gas: int = GAS_LIMIT) -> Enum:
+    def status(self) -> Enum:
         """Returns the status of the Job.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1003,9 +977,9 @@ class Job:
             Enum: returns the status as an enumeration.
 
         """
-        return status(self.job_contract, self.gas_payer)
+        return status(self.job_contract, self.gas_payer, self.gas)
 
-    def balance(self, gas: int = GAS_LIMIT) -> int:
+    def balance(self) -> int:
         """Retrieve the balance of a Job in HMT.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1032,7 +1006,7 @@ class Job:
 
         """
         return self.job_contract.functions.getBalance().call(
-            {"from": self.gas_payer, "gas": Wei(gas)}
+            {"from": self.gas_payer, "gas": Wei(self.gas)}
         )
 
     def manifest(self, priv_key: bytes) -> Dict:
@@ -1064,7 +1038,7 @@ class Job:
         """
         return download(self.manifest_url, priv_key)
 
-    def intermediate_results(self, priv_key: bytes, gas: int = GAS_LIMIT) -> Dict:
+    def intermediate_results(self, priv_key: bytes) -> Dict:
         """Reputation Oracle retrieves the intermediate results stored by the Recording Oracle.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1098,7 +1072,7 @@ class Job:
         """
         return download(self.intermediate_manifest_url, priv_key)
 
-    def final_results(self, priv_key: bytes, gas: int = GAS_LIMIT) -> Dict:
+    def final_results(self, priv_key: bytes) -> Dict:
         """Retrieves the final results stored by the Reputation Oracle.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1130,7 +1104,7 @@ class Job:
 
         """
         final_results_url = self.job_contract.functions.finalResultsUrl().call(
-            {"from": self.gas_payer, "gas": Wei(gas)}
+            {"from": self.gas_payer, "gas": Wei(self.gas)}
         )
         return download(final_results_url, priv_key)
 
@@ -1147,10 +1121,10 @@ class Job:
         gas_payer = credentials["gas_payer"]
         rep_oracle_priv_key = credentials["rep_oracle_priv_key"]
 
-        self.factory_contract = get_factory(factory_addr)
-        self.job_contract = get_escrow(escrow_addr)
-        self.manifest_url = manifest_url(self.job_contract, gas_payer)
-        self.manifest_hash = manifest_hash(self.job_contract, gas_payer)
+        self.factory_contract = get_factory(factory_addr, self.hmt_server_addr)
+        self.job_contract = get_escrow(escrow_addr, self.hmt_server_addr)
+        self.manifest_url = manifest_url(self.job_contract, gas_payer, self.gas)
+        self.manifest_hash = manifest_hash(self.job_contract, gas_payer, self.gas)
 
         manifest_dict = self.manifest(rep_oracle_priv_key)
         escrow_manifest = Manifest(manifest_dict)
@@ -1255,9 +1229,7 @@ class Job:
 
         return self._eth_addr_valid(gas_payer_addr, gas_payer_priv)
 
-    def _factory_contains_escrow(
-        self, escrow_addr: str, factory_addr: str, gas: int = GAS_LIMIT
-    ) -> bool:
+    def _factory_contains_escrow(self, escrow_addr: str, factory_addr: str) -> bool:
         """Checks whether a given factory address contains a given escrow address.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1290,16 +1262,15 @@ class Job:
             bool: returns True escrow belongs to the factory.
 
         """
-        factory_contract = get_factory(factory_addr)
+        factory_contract = get_factory(
+            factory_addr, hmt_server_addr=self.hmt_server_addr
+        )
         return factory_contract.functions.hasEscrow(escrow_addr).call(
-            {"from": self.gas_payer, "gas": Wei(gas)}
+            {"from": self.gas_payer, "gas": Wei(self.gas)}
         )
 
     def _init_factory(
-        self,
-        factory_addr: Optional[str],
-        credentials: Dict[str, str],
-        gas: int = GAS_LIMIT,
+        self, factory_addr: Optional[str], credentials: Dict[str, str]
     ) -> Contract:
         """Takes an optional factory address and returns its contract representation. Alternatively
         a new factory is created.
@@ -1333,16 +1304,21 @@ class Job:
         factory = None
 
         if not factory_addr_valid:
-            factory_addr = deploy_factory(GAS_LIMIT, **credentials)
-            factory = get_factory(factory_addr)
+            factory_addr = deploy_factory(
+                gas=self.gas, hmt_server_addr=self.hmt_server_addr, **credentials
+            )
+            factory = get_factory(factory_addr, hmt_server_addr=self.hmt_server_addr)
             if not factory_addr:
                 raise Exception("Unable to get address from factory")
 
         if not factory:
-            factory = get_factory(str(factory_addr))
+            factory = get_factory(
+                str(factory_addr), hmt_server_addr=self.hmt_server_addr
+            )
+
         return factory
 
-    def _bulk_paid(self, gas: int = GAS_LIMIT) -> int:
+    def _bulk_paid(self) -> int:
         """Checks if the last bulk payment has succeeded.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1376,10 +1352,10 @@ class Job:
 
         """
         return self.job_contract.functions.bulkPaid().call(
-            {"from": self.gas_payer, "gas": Wei(gas)}
+            {"from": self.gas_payer, "gas": Wei(self.gas)}
         )
 
-    def _last_escrow_addr(self, gas: int = GAS_LIMIT) -> str:
+    def _last_escrow_addr(self) -> str:
         """Gets the last deployed escrow contract address of the initialized factory contract.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1403,10 +1379,10 @@ class Job:
 
         """
         return self.factory_contract.functions.lastEscrow().call(
-            {"from": self.gas_payer, "gas": Wei(gas)}
+            {"from": self.gas_payer, "gas": Wei(self.gas)}
         )
 
-    def _create_escrow(self, trusted_handlers=[], gas: int = GAS_LIMIT) -> bool:
+    def _create_escrow(self, trusted_handlers=[]) -> bool:
         """Launches a new escrow contract to the ethereum network.
 
         >>> from test.hmt_escrow.utils import manifest
@@ -1443,7 +1419,8 @@ class Job:
         txn_info = {
             "gas_payer": self.gas_payer,
             "gas_payer_priv": self.gas_payer_priv,
-            "gas": gas,
+            "gas": self.gas,
+            "hmt_server_addr": self.hmt_server_addr,
         }
         func_args = [trusted_handlers]
 
@@ -1464,9 +1441,7 @@ class Job:
 
         return escrow_created
 
-    def _raffle_txn(
-        self, multi_creds, txn_func, txn_args, txn_event, gas: int = GAS_LIMIT
-    ):
+    def _raffle_txn(self, multi_creds, txn_func, txn_args, txn_event):
         """Takes in multiple credentials, loops through each and performs the given transaction.
 
         Args:
@@ -1486,7 +1461,8 @@ class Job:
             txn_info = {
                 "gas_payer": gas_payer,
                 "gas_payer_priv": gas_payer_priv,
-                "gas": gas,
+                "gas": self.gas,
+                "hmt_server_addr": self.hmt_server_addr,
             }
             try:
                 handle_transaction_with_retry(
